function varargout = hawkes_recover(evs,tmax,gamma,igamma,lagr_sparse,spconstr,gammalen)
% HAWKES_RECOVER - Find Max-Likelihood Parameters for Point Process
%  [A,MU,OBJ] = HAWKES_RECOVER(EV,TMAX,GAMMA,INTGAMMA,LAGR_SPARSE,SPCONSTR,GAMMALEN)
%  uses gradient descent to optimize the parameters A and MU for a Hawkes
%  process.  The infectivity parameters A are regularized using a sparsity
%  (L1) penalty with parameter LAGR_SPARSE.  The final objective is
%  returned in OBJ.
% 
%  The observed process EV can take one of two forms.  Either a structure
%  array with one struct per process and the field 'time' denoting event
%  times for that process, or a cell array where EV{1} is the number of
%  processes (an integer), EV{2} is a list of event times, and EV{3} is the
%  corresponding list of event process indices.
% 
%  The function handle GAMMA is the kernel used by the Hawkes process.  It
%  should take just one argument: time.  The function IGAMMA should be it's
%  integral up to a specified time.  Both are only evaluated for positive
%  arguments.
% 
%  These function handles can be replaced with cell arrays of function
%  handles.  In this case, A is Nx(N*K), where K is the number of kernels
%  given in GAMMA and IGAMMA.  Each block of N columns corresponds to the
%  relationship relative to one kernel, and reshape(A,[N,N,K]) will arrange
%  A into the familiar node-node relationships along the first two
%  dimensions and divide said slices by kernel in the third dimension.
% 
%  SPCONSTR allows one to impose a sparsity structure on the result. Zeros
%  in SPCONSTR will be zeros in the result, and nonzeros will be optimized
%  normally.
% 
%  GAMMALEN sets GAMMA(T)=0 for T>GAMMALEN, which can save computation.
%  When using multiple kernels, GAMMALEN should correspond to the longest
%  one.

narginchk(4,inf); % will automatically check the top limit, since not using varargin

numproc = evs{1};
evt0 = evs{2};
evp0 = evs{3}(:,1);
[evt,ind] = sort(evt0);
evp = evp0(ind);

npvec = (1:numproc)';

if nargin == 1 % useful for manually specified gmat/igsum
	if iscell(evs)
		evm = evs{3}(:,2:end);
	else
		evm = nan(numel(evt0),0);
	end
	pinds = arrayfun(@(i) {find(evp==i)},npvec); % get indices for each process
	varargout = {evt;evp;evm;pinds};
	return
end

if tmax < evt(end)
	warning('"TMAX" is less than maximum of supplied event times.')
end

if isa(gamma,'function_handle')
	gamma = {gamma};
elseif iscell(gamma) && all(cellfun(@(z) isa(z,'function_handle'),gamma));
else
	error('GAMMA must be a function handle or a cell array of function handles')
end
numg = numel(gamma);

if isa(igamma,'function_handle')
	igamma = {igamma};
elseif iscell(igamma) && numel(igamma) == numg && all(cellfun(@(z) isa(z,'function_handle'),igamma));
else
	error('IGAMMA must be a function handle or cell array of function handles (of the same size as GAMMA)')
end
igmat0 = nan(numproc,numg);
for i = 1:numg
	igmat0(1:numproc,i) = accumarray(evp,igamma{i}(tmax-evt),[numproc,1]);
end
igmat0 = reshape(igmat0,[numproc*numg,1]);

if nargin < 5 || isempty(lagr_sparse)
	lagr_sparse = zeros([numproc,numproc]);
elseif isscalar(lagr_sparse)
	lagr_sparse = repmat(lagr_sparse,[numproc,numproc]);
end
if nargin < 6 || isempty(spconstr)
	spconstr = true(numproc,numproc);
elseif any(size(spconstr) ~= numproc)
	error('SPCONSTR must be NxN, where N is the number of processes');
end
if nargin < 7 || isempty(gammalen)
	gammalen = inf;
end

MU = zeros(numproc,1);
A = zeros(numproc,numproc*numg);
obj0 = zeros(numproc,1);
gradcalcs = zeros(numproc,1);
llevals = zeros(numproc,1);
for i = 1:numproc
	pinds = find(evp==i);
	procnevs = numel(pinds);
	if procnevs == 0 % if this process has no activity, then the optimum is the all-zeros solution
		continue;
	end
	spconstri = spconstr(i,:); % get possible nonzeros for this row of A
	nonz = find(spconstri); % locations of possible nonzeros in row of A
	nnonz = numel(nonz);
	
	nonz0 = find(spconstri(evp)'); % events corresponding to possible nonzeros
	nonz0_rev = accumarray(nonz(:),1:nnonz,[numproc,1]); % to reverse the above
	[rsupp,csupp] = suppgmat(evt,pinds,gammalen,nonz0);
	
	gmat = nan(procnevs,nnonz*numg);
	for j = 1:numg
		val = gamma{j}(evt(pinds(rsupp))-evt(csupp));
		gmat(:,1+nnonz*(j-1):nnonz*j) = accumarray([rsupp,nonz0_rev(evp(csupp))],val,[procnevs,nnonz]);
	end
	nonz1 = repmat(nonz,[1,numg]);
	nonz2 = reshape(bsxfun(@plus,nonz',numproc*(0:numg-1)),[],1);
	igmat = igmat0(nonz1) + lagr_sparse(i,nonz1)';
	
	ma0 = [procnevs/tmax; zeros(nnonz*numg,1)]; % initialization parameters
	[ma,obj1,k,lleval0] = mlpoisson([ones(size(gmat,1),1),gmat],[tmax;igmat],1,ma0);
	
	MU(i) = ma(1);
	A(i,nonz2) = ma(2:end);
	obj0(i) = obj1;
	gradcalcs(i) = k;
	llevals(i) = lleval0;
end
obj = sum(obj0);
fprintf('Done in %i gradient and %i objective evaluations with objective %g\n',sum(gradcalcs),sum(llevals),obj)

varargout = {A;MU;obj};
end