function [evt,evm] = markpointproc(CIF,MKF,NUMEVTS,MARKDIM,HISTLEN)
% MARKPOINTPROC
%  [T,M] = MARKPOINTPROC(CIF,MKF,NUMEVTS,MARKDIM,HISTLEN) generates a
%  marked point process with NUMEVTS events using the conditional intensity
%  function CIF and mark generating function MKF.  It returns the times T
%  and marks M of the generated events, which respectively have dimensions 
%  NUMEVTSx1 and NUMEVTSxMARKDIM.
%  
%  The optional parameter HISTLEN can be used to limit the length (in time)
%  of history to use for CIF and MKF.  HISTLEN should be sufficiently large
%  to ensure that the effects of an event that far in the past are
%  negligible.  Default is HISTLEN=inf.
%  
%  The functions CIF and MKF should have the following formats:
% 
%  R = CIF(T,PASTT,PASTM) returns the CIF evaluated at time T, given past
%  event times PASTT and past marks PASTM.
%  M = MKF(T,PASTT,PASTM) returns a mark for an event at time T, given past
%  event times PASTT and past marks PASTM.  M should be a 1xMARKDIM vector.
%  
%  An alternative is:
%  [T,M] = MARKPOINTPROC(TGEN,MKF,NUMEVTS,MARKDIM,HISTLEN) uses TGEN in
%  place of CIF to generate event times.  TGEN should be defined
%  
%  T = TGEN(T0,PASTT,PASTM,TARG,GUESS) where T0 is the current time, PASTT
%  and PASTM are the past times and marks, and GUESS is the time between
%  the previous two generated events (which can make for a reasonable
%  initialization for iterative solvers, including the builtin solver).
%  This function should solve:
%  TARG == integral(@(t) CIF(t,PASTT,PASTM),T0,T) for T.
% 
%  In the CIF version of the function, the TGEN function is generated using
%  a customized solver (based closely on @quad).  While the CIF version is
%  general, custom TGEN functions can be much faster than the builtin.

% % for testing
% rng(8)
% n = 50;
% A = rand(n); A = A * 0.8/max(eig(A));
% mu = rand(n,1);
% A1 = sum(A,1);
% mu1 = sum(mu,1);
% % wrand = @(w) find(rand(1)*sum(w) < cumsum(w),1,'first'); % weighted random integer generation
% wrand = @(w) nnz([0;cumsum(w)]<rand(1)*sum(w)); % weighted random integer generation
% gamma = @(t) exp(-t);
% CIF = @(t,ht,hm) mu1 + A1(hm) * gamma(t-ht);
% MKF = @(t,ht,hm) wrand(mu + A(:,hm) * gamma(t-ht));
% NUMEVTS = 20000;
% MARKDIM = 1;

if nargin < 5 || isempty(HISTLEN)
	HISTLEN = inf;
end

evt = inf(NUMEVTS,1);
evm = nan(NUMEVTS,MARKDIM);

histstart = 1;
tcurr = 0;

if nargin(CIF) <= 3 % we have a CIF
	nexttime = @(tcurr,evtp,evmp,targ,guess) introot(@(t) CIF(t,evtp,evmp),tcurr+eps(tcurr),guess,targ,1e-6);
	tstep = 1/CIF(tcurr,evt([],1),evm([],:));
elseif nargin(CIF) > 3 % we have a function for finding the next event
	nexttime = CIF;
	tstep = 1;
end
% error = 0;
for i = 1:NUMEVTS
	targ = -log(rand);
	evtp = evt(histstart:i-1);
	evmp = evm(histstart:i-1,:);
	tnext = nexttime(tcurr,evtp,evmp,targ,tcurr+tstep);
	if(isnan(tnext) || tnext<=tcurr)
		save('markpointproc_error.mat')
		error('markpointproc() either failed to generate an event or generated one out-of-sequence (''earlier'' than an already generated event). It is likely that the process is unstable.  See <markpointproc3_error.mat> for more.')
	end
% 	error = error + abs(targ-quadv(@(t) CIF(t,evtp,evmp),tcurr,tnext));
	evt(i) = tnext;
	evm(i,:) = MKF(tnext,evtp,evmp);
	tstep = tnext-tcurr;
	tcurr = tnext;
	while tcurr-HISTLEN > evt(histstart)
		histstart = histstart+1;
	end
end

if nargout <= 1
	evt = [evt,evm];
end
end

function soln = introot(fun,a0,b0,targ0,tol)
% INTROOT - Find root of integration
%	R = INTROOT(FUN,A0,B0,T,TOL) attempts to find the R such that
%	
%	INTEGRAL(FUN,A0,R) == T
% 
%	Currently only works for T>0, R>A0, and possibly FUN(X)>0 for all X
%	Starts by looking on interval [A0,B0] with B0>A0, best if B0 is
%	slightly greater than R.  It does this using Simpson quadrature.

a = a0;
b = b0;
targ = targ0;

fa = fun(a);
fb = fun(b);
c = (a+b)/2;
fc = fun(c);
[y,soln] = myquad(fun,a,b,fa,fc,fb,tol,targ);
% args = [a,args,b];
% vals = [fa,vals,fb];

while isnan(soln)
	a = b;
	b = b+(b-a0);
	fa = fb;
	fb = fun(b);
	c = (a+b)/2;
	fc = fun(c);
	targ = targ-y;
	[y,soln] = myquad(fun,a,b,fa,fc,fb,tol,targ);
end
% fprintf('TARGET: %8.5g \tVALUE: %8.5g\n',targ0,quadv(fun,a0,soln,tol));

end

function [q,soln] = myquad(f,a,b,fa,fc,fb,tol,targ)
% Basically the inner loop of built-in quad() function, but without error
% checking

h = b-a;
c = (a+b)/2;
d = (a+c)/2;
e = (c+b)/2;

fd = f(d);
fe = f(e);

q1 = (h/6) * (fa + 4*fc + fb);
q2 = (h/12) * (fa + 4*fd + 2*fc + 4*fe + fb);

q = q2 + (q2-q1)/15;
if abs(q-q2) <= tol
	soln = nan;
	if q >= targ
% 		We should have some stop condition here
		quadpoly = [2,-4,2;-3,4,-1;1,0,0]*[fa;fc;fb]; % polynomial fit of points
		intpoly = [quadpoly./[3;2;1];-targ]; % integral of polynomial, offset by targ
		polyroots = real(roots(intpoly)); % solve the above polynomial
		soln = a + min(polyroots(polyroots>=0)); % find the closest solution on the interval
		if numel(soln) ~= 1 % something was messed up if we didn't get a single solution - debug to fix
			save('introot_error.mat')
			error('introot() needs fixing!  Check introot_error.mat')
		end
	end
	return
end

[qac,soln] = myquad(f,a,c,fa,fd,fc,tol/2,targ);
if ~isnan(soln) % we found the solution on this half
	return
end
[qcb,soln] = myquad(f,c,b,fc,fe,fb,tol/2,targ-qac);
q = qac+qcb;
end